"""
Offers publishing and subscribing to MQTT events.

Example use:

    from brewblox_service import mqtt, scheduler

    scheduler.setup(app)
    mqtt.setup(app)

    async def on_message(topic, body):
        print(f'Message published to {topic}')
        print(f'Message content: {body}')

    mqtt.listen(app, 'brewcast/state/a', on_message)
    mqtt.listen(app, 'brewcast/state/b', on_message)

    mqtt.subscribe(app, 'brewcast/state/#')

    await mqtt.publish('app', 'brewcast/state', {'example': True})
    await mqtt.publish('app', 'brewcast/state/a', {'example': True})
"""

import asyncio
import json
from contextlib import suppress
from dataclasses import dataclass, field
from ssl import CERT_NONE
from typing import Awaitable, Callable, List, Optional, Tuple, Union

import aiomqtt
from aiohttp import web
from aiohttp_apispec import docs, request_schema
from marshmallow import Schema, fields

from brewblox_service import brewblox_logger, features, strex

LOGGER = brewblox_logger(__name__)
routes = web.RouteTableDef()

EventData_ = Optional[Union[dict, list]]
ListenerCallback_ = Callable[[str, EventData_], Awaitable[None]]

DEFAULT_PORTS = {
    'mqtt': 1883,
    'mqtts': 8883,
    'ws': 80,
    'wss': 443,
}


def decoded(msg: Union[str, bytes, bytearray]):
    if isinstance(msg, (bytes, bytearray)):
        return msg.decode()
    return msg


@dataclass
class MQTTConfig:
    protocol: str
    host: str
    port: int
    path: str
    transport: str = field(init=False)
    client_will: dict = field(init=False)

    def __post_init__(self):
        if self.protocol not in ['ws', 'wss', 'mqtt', 'mqtts']:
            raise ValueError(f'Invalid protocol: {self.protocol}')

        if self.protocol.startswith('ws'):
            self.transport = 'websockets'
            self.path = self.path or ''
        else:
            self.transport = 'tcp'
            self.path = ''

        if not self.port or self.port == 5672:
            self.port = DEFAULT_PORTS[self.protocol]

        self.client_will = {}

    def __str__(self):
        return f'{self.protocol}://{self.host}:{self.port}{self.path}'


class EventHandler(features.ServiceFeature):
    """
    Connection handler class for MQTT events.
    Handles both TCP and Websocket connections.

    Automatically reconnects and resubscribes when connection is lost.

    Subscribe and listen are handled separately.
    You can subscribe to a wildcard topic, and then set callbacks for specific topics.

    Options are supplied either kwargs, or the service config.

    Kwargs:
        protocol ('ws' | 'wss' | 'mqtt' | 'mqtts'):
            Transport protocol used:
            - ws: websockets
            - wss: websockets over HTTPS
            - mqtt: TCP
            - mqtts: TCP + TLS/SSL

        host (str):
            Hostname of the broker.
            The default hostname of the RabbitMQ broker is 'eventbus'.
            This is accessible only from inside a Brewblox network.
            Outside of a Brewblox network, it is the hostname of the Traefik gateway.

        port (int):
            Port used by the broker.
            The default ports are:
            - ws: 80
            - wss': 443
            - mqtt': 1883
            - mqtts: 8883

            Note that this assumes ws/wss are connected to the Traefik gateway.
            If you are using websockets, and are directly connected to RabbitMQ, values are:
            - ws: 15675
            - wss: 15673

        path (str):
            Path to the broker.
            This value is ignored if mqtt/mqtts protocols are set.
            The default path for the RabbitMQ broker is '/eventbus'

            Examples: (formatted as <protocol>://<host>:<port><path>)
                ws://eventbus:15675/eventbus
                wss://BREWBLOX_HOST:443/eventbus
    """

    def __init__(self, app: web.Application, **args):
        super().__init__(app)

        config = app['config']
        protocol = args.get('protocol', config['mqtt_protocol'])
        host = args.get('host', config['mqtt_host'])
        port = args.get('port', config['mqtt_port'])
        path = args.get('path', config['mqtt_path'])
        self.config = MQTTConfig(protocol, host, port, path)
        self.client: aiomqtt.Client = None

        self._connect_ev: asyncio.Event = None
        self._subs: List[str] = []
        self._listeners: List[Tuple[str, ListenerCallback_]] = []

    def __str__(self):
        return f'<{type(self).__name__} for {self.config}>'

    @property
    def connected(self) -> bool:
        return self.client is not None \
            and self._connect_ev is not None \
            and self._connect_ev.is_set()

    def set_client_will(self, topic: str, message: EventData_, **kwargs):
        if self.client:
            raise RuntimeError('Client will must be set before startup')
        payload = json.dumps(message) if message is not None else None
        self.config.client_will = dict(topic=topic,
                                       payload=payload,
                                       **kwargs)

    @staticmethod
    def create_client(config: MQTTConfig) -> aiomqtt.Client:
        client = aiomqtt.Client(transport=config.transport,
                                protocol=aiomqtt.MQTTv311)
        client.ws_set_options(path=config.path)

        if config.protocol in ['mqtts', 'wss']:
            client.tls_set(cert_reqs=CERT_NONE)
            client.tls_insecure_set(True)

        if config.client_will:
            client.will_set(**config.client_will)

        return client

    async def startup(self, app: web.Application):
        self._connect_ev = asyncio.Event()
        self.client = self.create_client(self.config)
        self.client.on_connect = self._on_connect
        self.client.on_disconnect = self._on_disconnect
        self.client.on_message = self._on_message

        LOGGER.info(f'Starting {self}')
        self.client.connect_async(self.config.host, self.config.port)
        self.client.loop_start()

    async def shutdown(self, app: web.Application):
        if self.client:
            self.client.disconnect()
            await self.client.loop_stop()
            self.client = None

    def _on_connect(self, *args):
        LOGGER.debug(f'Applying subscribe for {self._subs}')
        for topic in self._subs:
            self.client.subscribe(topic)

        LOGGER.info(f'{self} connected')
        self._connect_ev.set()

    def _on_disconnect(self, *args):
        LOGGER.info(f'{self} disconnected')
        self._connect_ev.clear()

    async def _handle_callback(self, cb, topic, payload):
        try:
            await cb(topic, json.loads(payload) if payload else None)
        except Exception as ex:
            LOGGER.error(f'Exception handling MQTT callback for {topic}: {strex(ex)}')

    def _on_message(self, client, userdata, message, *args):
        try:
            topic = decoded(message.topic)
            payload = decoded(message.payload)
        except UnicodeDecodeError:  # pragma: no cover
            LOGGER.error('Skipping malformed MQTT event')
            return

        matching = [cb
                    for (sub, cb) in self._listeners
                    if aiomqtt.topic_matches_sub(sub, topic)]

        for cb in matching:
            asyncio.create_task(self._handle_callback(cb, topic, payload))

        if not matching:
            LOGGER.info(f'{self} recv topic={topic}, msg={payload[:30]}...')

    async def publish(self, topic: str, message: EventData_, err=True, **kwargs):
        payload = json.dumps(message) if message is not None else None
        info = self.client.publish(topic, payload, **kwargs)
        error = aiomqtt.error_string(info.rc)
        LOGGER.debug(f'publish({topic}) -> {error}')
        if info.rc != 0 and err:
            raise ConnectionError(f'Publish error="{error}", topic="{topic}"')

    async def subscribe(self, topic: str):
        LOGGER.info(f'subscribe({topic})')
        self._subs.append(topic)
        if self.connected:
            self.client.subscribe(topic)

    async def listen(self, topic: str, callback: ListenerCallback_):
        LOGGER.info(f'listen({topic})')
        self._listeners.append((topic, callback))

    async def unsubscribe(self, topic: str):
        LOGGER.info(f'unsubscribe({topic})')
        if self.connected:
            self.client.unsubscribe(topic)
        with suppress(ValueError):
            self._subs.remove(topic)

    async def unlisten(self, topic: str, callback: ListenerCallback_):
        LOGGER.info(f'unlisten({topic})')
        with suppress(ValueError):
            self._listeners.remove((topic, callback))


def setup(app: web.Application):
    """
    Initializes the EventHandler in the app context.

    Args:
        app (web.Application):
            The Aiohttp Application object.
    """
    features.add(app, EventHandler(app))
    app.router.add_routes(routes)


def handler(app: web.Application) -> EventHandler:
    """
    Get registered EventHandler.
    Requires setup(app) to have been called first.

    Args:
        app (web.Application):
            The Aiohttp Application object.
    """
    return features.get(app, EventHandler)


def set_client_will(app: web.Application,
                    topic: str,
                    message: EventData_ = None,
                    **kwargs):
    """
    Set MQTT Last Will and Testament for client.
    Requires setup(app) to have been called first.
    Must be called before startup.

    Args:
        app (web.Application):
            The Aiohttp Application object.

        topic (str):
            The MQTT message topic. Cannot include wildcards.

        message (dict, None):
            A JSON-serializable object, or None.
    """
    handler(app).set_client_will(topic, message, **kwargs)


async def publish(app: web.Application,
                  topic: str,
                  message: EventData_,
                  **kwargs):
    """
    Publish a new event message.

    Shortcut for `handler(app).subscribe(topic, message)`.
    Requires setup(app) to have been called first.

    Args:
        app (web.Application):
            The Aiohttp Application object.

        topic (str):
            The MQTT message topic. Cannot include wildcards.

        message (dict):
            A JSON-serializable object.
    """
    await handler(app).publish(topic, message, **kwargs)


async def subscribe(app: web.Application, topic: str):
    """
    Subscribe to event messages.
    Requires setup(app) to have been called first.

    In order to get callbacks for events, listen() must also be used.
    You can register multiple listeners for a single subscribed topic.

    See: http://www.steves-internet-guide.com/understanding-mqtt-topics/

    Args:
        app (web.Application):
            The Aiohttp Application object.

        topic (str):
            A filter for message topics.
            Can include the '+' and '#' wildcards.
    """
    await handler(app).subscribe(topic)


async def listen(app: web.Application, topic: str, callback: ListenerCallback_):
    """
    Set a listener for event messages.
    Requires setup(app) to have been called first.

    In order to get callbacks, subscribe() must also be used.
    You can register multiple listeners for a single subscribed topic.

    See: http://www.steves-internet-guide.com/understanding-mqtt-topics/

    Args:
        app (web.Application):
            The Aiohttp Application object.

        topic (str):
            A filter for message topics.
            Can include the '+' and '#' wildcards.

    """
    await handler(app).listen(topic, callback)


async def unsubscribe(app: web.Application, topic: str):
    """
    Unsubscribe to event messages.
    Requires setup(app) to have been called first.

    Removes a subscription that was set by `subscribe(topic)`.
    Does nothing if no subscription can be found.

    Args:
        app (web.Application):
            The Aiohttp Application object.

        topic (str):
            Must match the `topic` argument earlier used in `subscribe(topic)`.
    """
    await handler(app).unsubscribe(topic)


async def unlisten(app: web.Application, topic: str, callback: ListenerCallback_):
    """
    Remove a callback for received event messages.
    Requires setup(app) to have been called first.

    Removes a listener that was set by `listen(topic, callback)`.
    Both `topic` and `callback` must match for the listener to be removed.
    Does nothing if no listener can be found found.

    Args:
        app (web.Application):
            The Aiohttp Application object.

        topic (str):
            Must match the `topic` argument earlier used in `listen(topic, callback)`.

        callback (ListenerCallback_):
            Must match the `callback` argument earlier used in `listen(topic, callback)`.
    """
    await handler(app).unlisten(topic, callback)


class MQTTPublishSchema(Schema):
    topic = fields.String()
    message = fields.Dict()


class MQTTSubscribeSchema(Schema):
    topic = fields.String()


@docs(
    tags=['MQTT'],
    summary='Publish an event message.',
    description='This is a debugging / diagnostics endpoint.'
)
@routes.post('/_debug/publish')
@request_schema(MQTTPublishSchema())
async def post_publish(request):
    data = request['data']
    await publish(request.app, data['topic'], data['message'])
    return web.Response()


@docs(
    tags=['MQTT'],
    summary='Subscribe to event messages.',
    description='This is a debugging / diagnostics endpoint. '
    'Messages received for this subscription will be logged and then discarded.'
)
@routes.post('/_debug/subscribe')
@request_schema(MQTTSubscribeSchema())
async def post_subscribe(request):
    await subscribe(request.app, request['data']['topic'])
    return web.Response()



"""
Testing utility functions
"""

import re

from aiohttp.client_exceptions import ContentTypeError


async def response(request, status=200):
    retv = await request
    if retv.status != status:
        print(retv)
        assert retv == status
    try:
        return await retv.json()
    except ContentTypeError:
        return await retv.text()


class matching:
    """Assert that a given string meets some expectations."""

    def __init__(self, pattern, flags=0):
        self._regex = re.compile(pattern, flags)

    def __eq__(self, actual):
        return bool(self._regex.match(actual))

    def __repr__(self):
        return self._regex.pattern



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['brewblox_service']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp-apispec>=2.2.1,<3.0.0',
 'aiohttp>=3.7.3,<4.0.0',
 'aiomqtt>=0.1.1,<0.2.0']

setup_kwargs = {
    'name': 'brewblox-service',
    'version': '0.32.2',
    'description': 'Scaffolding for Brewblox backend services',
    'long_description': "# Scaffolding for Brewblox service applications\n\nIn order to reduce code duplication between services, generic functionality is implemented here.\n\nFor an example on how to implement your own service based on `brewblox-service`, see <https://github.com/brewblox/brewblox-boilerplate>.\n\n`brewblox-service` can technically be launched as a standalone application, but will not be very useful.\n\n## [brewblox_service](./brewblox_service/__init__.py)\n\nSmall generic tools are defined here.\n\n`brewblox_logger` can be used for creating module-specific loggers. It is not required, but will play a bit nicer with default formatting of the log.\n\nExample:\n\n```python\nfrom brewblox_service import brewblox_logger\n\nLOGGER = brewblox_logger(__name__)\nLOGGER.info('hello')\n```\n\n## [service.py](./brewblox_service/service.py)\n\nParses commandline arguments, creates an `aiohttp` app, and runs it.\n\nThe shortest implementation is:\n\n```python\napp = service.create_app(default_name='my_service')\nservice.furnish(app)\nservice.run(app)\n```\n\nThis will get you a working web application, but it will only support the `/_service/status` health check endpoint.\n\nApplications can configure their own features, and add new commandline arguments.\n\nExample:\n\n```python\n# Separately creating the parser allows adding arguments to the default set\nparser = service.create_parser(default_name='my_service')\nparser.add_argument('--my-arg')\n\n# Now create the app\napp = service.create_app(parser=create_parser())\n\n# Add features for this service\ndevice.setup(app)\napi.setup(app)\n\n# Furnish and run\nservice.furnish(app)\nservice.run(app)\n```\n\n## [features.py](./brewblox_service/features.py)\n\nMany service features are application-scoped. Their lifecycle should span multiple requests, either because they are not request-driven, or because they manage asynchronous I/O operations (such as listening to AMQP messages).\n\nThe `ServiceFeature` class offers an abstract base class for this behavior. Implementing classes should define `startup(app)` and `shutdown(app)` functions, and those will be automatically called when the application starts up and shuts down.\n\nBoth `startup()` and `shutdown()` are called in an async context, making them the async counterparts of `__init__()` and `__del__()` functions.\n\nFeatures must be constructed after the app is created, but before it starts running. (`service.create_app()` and `service.run(app)`)\n\nThe `add()` and `get()` functions make it easy to centrally declare a feature, and then use it in any function that has a reference to the aiohttp app.\n\n## [mqtt.py](./brewblox_service/mqtt.py)\n\nThe replacement for the AMQP-based `events` module.\n\nThe major advantages over AMQP is that the protocol is simpler, and supports websocket transport.\nThis allows for it to be used outside a Brewblox network.\n\nMessages have a /-separated *topic* and a JSON-serialized body.\n\nTo publish data, all you need is the `publish(topic, message)` function.\n\nTo listen to incoming messages, you can combine `subscribe(topic)` with one or more calls to `listen(topic, callback)`.\nThe subscribe/listen functions allow for + and # wildcards to be used.\n\nFor a detailed explanation of how to use MQTT topics, see http://www.steves-internet-guide.com/understanding-mqtt-topics/.\n\nFor the Brewblox spec on how and where to publish data, see https://brewblox.netlify.app/dev/reference/event_logging.html.\n\nIncludes top-level convenience functions for `publish(topic, message)`, `listen(topic, callback)` and `subscribe(topic)`.\n",
    'author': 'BrewPi',
    'author_email': 'development@brewpi.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



"""
Generic startup functions for a brewblox application.

Responsible for parsing user configuration, and creating top-level objects.
This module provides the framework to which service implementations can attach their features.

Example:
    # Uses the default argument parser
    # To add new commandline arguments, use create_parser()
    app = service.create_app(default_name='my_service')

    # (Placeholder names)
    # All features (endpoints and async handlers) must be created and added to the app here
    # The Aiohttp Application will freeze functionality once it has been started
    feature_one.setup(app)
    feature_two.setup(app)

    # Modify added resources to conform to standards
    service.furnish(app)

    # Run the application
    # This function blocks until the application is shut down
    service.run(app)
"""

import argparse
import logging
# The argumentparser can't fall back to the default sys.argv if sys is not imported
import sys  # noqa
import tempfile
from os import getenv
from typing import List, Optional

from aiohttp import web
from aiohttp_apispec import docs, setup_aiohttp_apispec, validation_middleware

from brewblox_service import brewblox_logger, cors, features

LOGGER = brewblox_logger(__name__)
routes = web.RouteTableDef()


def _init_logging(args: argparse.Namespace):
    level = logging.DEBUG if args.debug else logging.INFO
    format = '%(asctime)s %(levelname)-8s %(name)-30s  %(message)s'
    datefmt = '%Y/%m/%d %H:%M:%S'

    logging.basicConfig(level=level, format=format, datefmt=datefmt)
    logging.captureWarnings(True)

    if not args.debug:
        logging.getLogger('aiohttp.access').setLevel(logging.WARN)


def create_parser(default_name: str) -> argparse.ArgumentParser:
    """
    Creates the default brewblox_service ArgumentParser.
    Service-agnostic arguments are added.

    The parser allows calling code to add additional arguments before using it in create_app()

    Args:
        default_name (str):
            default value for the --name commandline argument.

    Returns:
        argparse.ArgumentParser: a Python ArgumentParser with defaults set.

    """
    parser = argparse.ArgumentParser(fromfile_prefix_chars='@')
    parser.add_argument('-H', '--host',
                        help='Host to which the app binds. [%(default)s]',
                        default='0.0.0.0')
    parser.add_argument('-p', '--port',
                        help='Port to which the app binds. [%(default)s]',
                        default=5000,
                        type=int)
    parser.add_argument('-n', '--name',
                        help='Service name. This will be used as prefix for all endpoints. [%(default)s]',
                        default=default_name)
    parser.add_argument('--debug',
                        help='Run the app in debug mode. [%(default)s]',
                        action='store_true')

    group = parser.add_argument_group('MQTT Event handling')
    group.add_argument('--mqtt-protocol',
                       help='Transport protocol used for MQTT events. [%(default)s]',
                       choices=['mqtt', 'mqtts', 'ws', 'wss'],
                       default='mqtt')
    group.add_argument('--mqtt-host',
                       help='Hostname at which the eventbus can be reached [%(default)s]',
                       default='eventbus')
    group.add_argument('--mqtt-port',
                       help='Port at which the eventbus can be reached [%(default)s]',
                       type=int)
    group.add_argument('--mqtt-path',
                       help='Path used for MQTT events. Only applies if a websockets protocol is used. [%(default)s]',
                       default='/eventbus')
    group.add_argument('--history-topic',
                       help='Eventbus exchange to which logged controller state is broadcast. [%(default)s]',
                       default='brewcast/history')
    group.add_argument('--state-topic',
                       help='Eventbus topic to which volatile controller state is broadcast. [%(default)s]',
                       default='brewcast/state')
    return parser


def create_app(
        default_name: str = None,
        parser: argparse.ArgumentParser = None,
        raw_args: Optional[List[str]] = None
) -> web.Application:
    """
    Creates and configures an Aiohttp application.

    Args:
        default_name (str, optional):
            Default value for the --name commandline argument.
            This value is required if `parser` is not provided.
            This value will be ignored if `parser` is provided.

        parser (argparse.ArgumentParser, optional):
            Application-specific parser.
            If not provided, the return value of `create_parser()` will be used.

        raw_args (list of str, optional):
            Explicit commandline arguments.
            Defaults to sys.argv[1:]

    Returns:
        web.Application: A configured Aiohttp Application object.
            This Application must be furnished, and is not yet running.

    """

    if parser is None:
        assert default_name, 'Default service name is required'
        parser = create_parser(default_name)

    args, unknown_args = parser.parse_known_args(raw_args)
    _init_logging(args)

    if unknown_args:
        LOGGER.error(f'Unknown arguments detected: {unknown_args}')

    app = web.Application()
    app['config'] = vars(args)
    prefix = '/' + args.name.lstrip('/')

    app.middlewares.append(cors.cors_middleware)
    app.middlewares.append(validation_middleware)

    setup_aiohttp_apispec(
        app=app,
        title=args.name,
        version='v1',
        url=f'{prefix}/api/doc/swagger.json',
        swagger_path=f'{prefix}/api/doc',
        static_path=f'{prefix}/static/swagger',
    )

    return app


def furnish(app: web.Application):
    """
    Configures Application routes, readying it for running.

    This function modifies routes and resources that were added by calling code,
    and must be called immediately prior to `run(app)`.

    Args:
        app (web.Application):
            The Aiohttp Application as created by `create_app()`
    """
    config = app['config']
    name = config['name']
    prefix = '/' + name.lstrip('/')

    prefixed_paths = [
        f'{prefix}/api/doc/swagger.json',
        f'{prefix}/api/doc',
        f'{prefix}/static/swagger',
    ]

    app.router.add_routes(routes)
    for resource in app.router.resources():
        if resource.canonical not in prefixed_paths:
            resource.add_prefix(prefix)

    LOGGER.info(f'Service name: {name}')
    LOGGER.info(f'Service info: {getenv("SERVICE_INFO")}')
    LOGGER.info(f'Service config: {config}')

    for route in app.router.routes():
        LOGGER.debug(f'Endpoint [{route.method}] {route.resource.canonical}')

    for name, impl in app.get(features.FEATURES_KEY, {}).items():
        LOGGER.debug(f'Feature [{name}] {impl}')


def run(app: web.Application, listen_http: bool = True):
    """
    Runs the application in an async context.
    This function will block indefinitely until the application is shut down.

    Args:
        app (web.Application):
            The Aiohttp Application as created by `create_app()`

        listen_http (bool):
            Whether to open a port for the REST API.
            Set to False to disable all REST endpoints.
            This can be useful for services that use communication protocols
            other than REST (such as MQTT), or only have active functionality.
    """
    config = app['config']

    if listen_http:
        web.run_app(app, host=config['host'], port=config['port'])
    else:
        # Listen to a dummy UNIX socket
        # The service still runs, but is not bound to any port
        # This is useful for services without a meaningful REST API
        with tempfile.TemporaryDirectory() as tmpdir:
            web.run_app(app, path=f'{tmpdir}/dummy.sock')


@docs(
    tags=['Service'],
    summary='Service health check',
)
@routes.get('/_service/status')
async def healthcheck(request: web.Request) -> web.Response:
    return web.json_response({'status': 'ok'})
